html,
body {
	position: relative;
	width: 100%;
	height: 100%;
}

body {
	color: var(--G1);
	margin: 0;
	padding: 0;
	box-sizing: border-box;
	font-family: "Open Sans", Roboto, Oxygen-Sans, Ubuntu, Cantarell,
		"Helvetica Neue", sans-serif;
	font-size: 14px;
}

.autocomplete-input {
	min-width: 100px;
	background: none;
	border-radius: 5px;
	border: 1px solid var(--G4);
}

.autocomplete {
	min-width: 100px !important;
}

p {
	font-size: 14px;
	color: var(--G2);
}

h4 {
	font-size: 14px;
	font-weight: 600;
	color: var(--G2);
}

h2 {
	font-weight: 500;
}

small {
	font-size: 10px;
	color: var(--G2);
}
.mdc-data-table__pagination {
	border-top-width: 0px;
}

table {
	overflow: visible;
}

.end-cell {
	border-right: 1px solid var(--G5);
}

.remove {
	cursor: pointer;
}
